//  Fish Supper
//  Copyright (C) 2006 Matthew Clarke <matthew_larcke@lycos.co.uk>
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//  02111-1307, USA.




#include "spritemanager.h"
#include <fstream>
#include <cstdio>
#include <iostream>
#include <cstdlib>

using std::ifstream;
using std::cerr;
using std::cout;
using std::endl;




// *******************
// *** CONSTRUCTOR ***
// *******************
FS::SpriteManager::SpriteManager()
{
    // WARNING: hard-coded!
    int temp_min_ys[] = { 87, 162, 237, 312, 387, 462 };
    int temp_max_ys[] = { 136, 211, 286, 361, 436, 511 };
     
    for (int i = 0; i < NUM_ROWS; ++i)
    {
        rows_min_ys[i] = temp_min_ys[i];
        rows_max_ys[i] = temp_max_ys[i];
    } // for

    my_logs_allocated = false;
    my_crystals_allocated = false;
    
} // FS::SpriteManager::SpriteManager()

// ******************
// *** DESTRUCTOR ***
// ******************
FS::SpriteManager::~SpriteManager() 
{
    if (my_logs_allocated)
    {
        for (int i = 0; i < NUM_ROWS; ++i)
        {
            delete [] my_logs[i];
        } // for  
    } // if
    
    if (my_crystals_allocated)
    {
        delete [] my_crystals;
    } // if
      
} // FS::SpriteManager::~SpriteManager() 




// ************************
// *** MEMBER FUNCTIONS ***
// ************************

// **************************************************

void FS::SpriteManager::create_logs(const int * row_lengths, const Block * const * my_blocks,
        const int * row_lengths_pixels, const Direction * directions)
{
    int start_index;
    int end_index;
    LogColor color;
    int log_count;
    
    // allocate memory for my_logs arrays
    // call delete [] if necessary
    for (int r = 0; r < NUM_ROWS; ++r)
    {
        if (my_logs_allocated)
        {
            delete [] my_logs[r];
        } // if
        my_logs[r] = new Log[num_logs[r]];
        
        // reset indices
        start_index = end_index = -1;
        log_count = 0;
        
        for (int c = 0; c < row_lengths[r]; ++c)
        {
            if (my_blocks[r][c].lc != L_BACKGROUND)
            {
                if (start_index < 0)
                {
                    start_index = c;
                    color = my_blocks[r][c].lc;
                } // if
            }
            else if (start_index >= 0)  // current block is background
            {
                end_index = c;

                my_logs[r][log_count].init(rows_min_ys[r],
                                           (end_index - start_index) * BLOCK_WIDTH,
                                           BLOCK_HEIGHT, color, start_index * BLOCK_WIDTH,
                                           row_lengths_pixels[r], directions[r], r);
                
                start_index = end_index = -1;
                ++log_count;
            } // if ... else
        } // for c
    } // for r
    
    my_logs_allocated = true;
    
} // FS::SpriteManager::create_logs()

// **************************************************

void FS::SpriteManager::create_crystals(const int * row_lengths, 
        const Block * const * my_blocks, const int * row_lengths_pixels, 
        const Direction * directions)
{
    // allocate memory for Crystals 
    // delete previous contents of my_crystals if necessary
    if (my_crystals_allocated)
    {
        delete [] my_crystals;
    } // if
    
    my_crystals = new Crystal[num_crystals];
    
    // find the crystals and initialize them
    int crystal_count = 0;
    //LogColor erase_color;
    bool at_front, at_back;
    
    for (int r = 0; r < NUM_ROWS; ++r)
    {
        for (int c = 0; c < row_lengths[r]; ++c)
        {
            if (my_blocks[r][c].cc != C_NONE)
            {
                /*
                // What's the previous log color, given this row's direction?
                // That's the crystals erase_color. Assume that no crystal will
                // be at the edge of a row (i.e. column is > 0 and < length-1)...
                // OTHERWISE IT BREAKS!!!
                switch (directions[r])
                {
                    case LEFT_TO_RIGHT:
                        erase_color = my_blocks[r][c-1].lc;
                        break;
                    case RIGHT_TO_LEFT:
                        erase_color = my_blocks[r][c+1].lc;
                        break;
                } // switch
                cout << "erase_color=" << ((int) erase_color) << endl;
                */
                       
                // WARNING! Following code breaks if a crystal is placed
                // at far end of map (i.e. at 0 or row_length-1)!!!
                switch (directions[r])
                {
                    case LEFT_TO_RIGHT:
                        at_front = (my_blocks[r][c+1].lc == L_BACKGROUND);
                        at_back = (my_blocks[r][c-1].lc == L_BACKGROUND);
                        break;
                    case RIGHT_TO_LEFT:
                        at_front = (my_blocks[r][c-1].lc == L_BACKGROUND);
                        at_back = (my_blocks[r][c+1].lc == L_BACKGROUND);
                        break;
                } // switch
                        
                my_crystals[crystal_count].init(c * BLOCK_WIDTH,    
                        r, rows_min_ys[r], BLOCK_WIDTH, BLOCK_HEIGHT, 
                        my_blocks[r][c].cc, at_front, at_back,
                        row_lengths_pixels[r], directions[r]);
                ++crystal_count;
            } // if
        } // for
    } // for
                
} // FS::SpriteManager::create_crystals()

// **************************************************

void FS::SpriteManager::reset()
{
    the_cat.reset();    
    is_first_pass = true;
    cat_crystal_collisions_disabled = false;
    disabled_crystal = 0;
    
    current_state = PLAYING_GAME;
    
    my_snd_engine->play_sound(SoundEngine::RIVER, 1, -1, 30);
        
} // FS::SpriteManager::reset()
         
// **************************************************

void FS::SpriteManager::restart_level()
{
    the_cat.reset();
    
    // These two shouldn't need resetting, but won't do any
    // harm to include these lines anyway...
    cat_crystal_collisions_disabled = false;
    disabled_crystal = 0;
    
    for (int i = 0; i < num_crystals; ++i)
    {
        my_crystals[i].reset();
    } // for
    
    //my_player_data->draw_crystals_held(my_play_display);
    
    redraw_crystals = true;   // forces crystals held to be re-drawn
    
    current_state = PLAYING_GAME;
    
} // FS::SpriteManager::restart_level()

// **************************************************

FS::SpriteManager::State FS::SpriteManager::update(int time_passed, KeyboardEvent ke,
        const int * current_start_pixels, const int * pixels_to_move, 
        const int * old_pixels_to_move)
{
    update_logs_and_crystals(current_start_pixels, pixels_to_move, old_pixels_to_move);

    // FIXME - a bit of a hack!
    // N.B. This'll never happen on first pass.
    if ( the_cat.get_y() < UPPER_BANK_MAX_Y )
    {
        the_fish.draw(my_play_display);
    } // if
    
    the_cat.update(time_passed, ke, my_play_display);

    // There won't be any collisions on first pass anyway, but
    // crystals need drawing now.
    if (is_first_pass)
    {
        my_player_data->draw_crystals_held(my_play_display);
        the_fish.draw(my_play_display);
        my_play_display->request_draw_score(my_player_data->get_score());
        my_play_display->request_draw_lives(my_player_data->get_num_lives());
        my_play_display->request_draw_level(my_player_data->get_level());
    } // if
     
    switch (current_state)
    {
        case PLAYING_GAME:
            // If Albert is airborne or in water, we don't need
            // to check for any collisions.
            if ( the_cat.is_airborne() )
            {
                break;
            }
            
            if ( the_cat.is_in_water_or_being_zapped() )
            {
                if ( the_cat.is_splashed_zapped_out() )
                {
                    my_player_data->dec_num_lives();
                    my_play_display->request_draw_lives(my_player_data->get_num_lives());
                    if ( !my_player_data->get_num_lives() )
                    {
                        the_cat.start_leave_in_a_huff(time_passed);
                        current_state = RUNNING_GAME_OVER_ROUTINE;
                        my_snd_engine->play_sound(SoundEngine::MIAOW);
                        break;
                    } 
                    else
                    {
                        current_state = REQUEST_LEVEL_RESTART;
                        break;
                    } // if ... else
                } // if
                break;  // don't want collisions checking! 
            } // if 
            
            check_collisions(time_passed);
            
            break;
        
        case RUNNING_LEVEL_COMPLETED_ROUTINE:
            if ( the_cat.has_left() )
            {
                current_state = LEVEL_COMPLETED;
                my_snd_engine->stop_sound(1);
            } // if
            break;
            
        case RUNNING_GAME_OVER_ROUTINE:
            if ( the_cat.has_left() )
            {
                current_state = GAME_OVER;
                my_snd_engine->stop_sound(1);
            } // if
            break;
        
        case REQUEST_LEVEL_RESTART:    
        case LEVEL_COMPLETED:
        case GAME_OVER:
            break;
            
    } // switch
    
    if ( !is_first_pass )
    {
        if ( the_cat.get_y() < UPPER_BANK_MAX_Y )
        {
            my_play_display->request_draw_score(my_player_data->get_score());
            my_play_display->request_draw_lives(my_player_data->get_num_lives());
            my_play_display->request_draw_level(my_player_data->get_level());
        } // if
        
        if ( redraw_crystals || (the_cat.get_y() >= (LOWER_BANK_MIN_Y - ALBERT_HEIGHT)) )
        {
            my_player_data->draw_crystals_held(my_play_display);
        } // if    
    } // if    
        
    is_first_pass = false;
    redraw_crystals = false;
    
    return current_state;
    
} // FS::SpriteManager::update()
    
// **************************************************

void FS::SpriteManager::check_collisions(int time_passed)
{    
    if ( the_cat.is_on_log() )
    {
        check_cat_crystal_collision( the_cat.get_host_log_row() );
    }
    else
    {
        RowDescription row_desc = calc_alberts_row();
        
        switch (row_desc)
        {
            case ROW_0:
            case ROW_1:
            case ROW_2:
            case ROW_3:
            case ROW_4:
            case ROW_5: 
                // check logs on given row
                // N.B. Check for crystals first and Albert would be able
                // to grab a crystal just before he landed on a log - might be
                // the crystal that'd keep him on that log!
                switch ( check_cat_log_collision((int) row_desc) )
                {
                    case ALLOWED:
                        break;
                    case DISALLOWED:
                        the_cat.start_zap(time_passed);
                        break;
                    case MISSED:
                        the_cat.start_splash(time_passed);
                        my_snd_engine->play_sound(SoundEngine::SPLASH);
                        break;
                } // switch
                break;
                
            case UPPER_BANK:
                // draw score to make sure Albert doesn't erase it!
                my_play_display->request_draw_score(my_player_data->get_score());
                my_play_display->request_draw_lives(my_player_data->get_num_lives());
                
                // check for collision with fish
                if ( /*check_cat_fish_collision()*/ the_cat.check_fish(&the_fish) )
                {
                    my_player_data->inc_score(PTS_FISH);
                    my_play_display->request_draw_score(my_player_data->get_score());
                    the_cat.start_leave_with_fish(time_passed);
                    current_state = RUNNING_LEVEL_COMPLETED_ROUTINE;
                } // if
                break;
                
            case LOWER_BANK:
                // draw crystals held to make sure Albert doesn't erase them!
                //my_player_data->draw_crystals_held(my_play_display);
                break;
                
            case WATER:
                // SPLASH!!!
                the_cat.start_splash(time_passed);
                my_snd_engine->play_sound(SoundEngine::SPLASH);
                break;
                
        } // switch
        
    } // if ... else
      
} // FS::SpriteManager::check_collisions() 
        
// **************************************************

FS::RowDescription FS::SpriteManager::calc_alberts_row()
{
    if ( the_cat.get_collision_box_y1() >= LOWER_BANK_MIN_Y )
    {
        return LOWER_BANK;
    }
    else if ( the_cat.get_collision_box_y2() <= UPPER_BANK_MAX_Y )
    {
        return UPPER_BANK;
    }
    else
    {
        for (int i = 0; i < NUM_ROWS; ++i)
        {
            if ( (the_cat.get_collision_box_y1() >= rows_min_ys[i]) &&
                    (the_cat.get_collision_box_y2() <= rows_max_ys[i]) )
            {
                return (RowDescription) i;
            } // if
        } // for
    } // if ... else
    
    return WATER;
    
} // FS::SpriteManager::calc_alberts_row()
            
// **************************************************

FS::SpriteManager::CatLogCollisionType FS::SpriteManager::check_cat_log_collision(int row)
{
    for (int i = 0; i < num_logs[row]; ++i)
    {
        if ( the_cat.check_new_log( & my_logs[row][i] ) )
        {
            // Albert class will deal with setting host_log, etc...
            if ( my_player_data->is_allowed_on_log(my_logs[row][i].get_color(), 
                    my_play_display) )
            {
                // add points
                switch ( my_logs[row][i].get_color() )
                {
                    case L_RED:
                    case L_YELLOW:
                    case L_BLUE:
                        my_player_data->inc_score(PTS_ONE_COLOR_LOG);
                        my_play_display->request_draw_score(my_player_data->get_score());
                        break;
                    case L_ORANGE:
                    case L_GREEN:
                    case L_PURPLE:
                        my_player_data->inc_score(PTS_TWO_COLOR_LOG);
                        my_play_display->request_draw_score(my_player_data->get_score());
                        break;
                    case L_BROWN:
                        my_player_data->inc_score(PTS_BROWN_LOG);
                        my_play_display->request_draw_score(my_player_data->get_score());
                        break;
                    case L_BACKGROUND:  // should never happen
                        break;
                } // switch
                redraw_crystals = true;
                return ALLOWED;
            }
            else
            {
                return DISALLOWED;
            } // if ... else
        } // if
    } // for
    
    return MISSED;   // i.e. SPLASH!!!
    
} // FS::SpriteManager::check_cat_log_collision()

// **************************************************

void FS::SpriteManager::check_cat_crystal_collision(int row)
{
    if (cat_crystal_collisions_disabled)
    {
        // Is Albert still 'over' the disabled crystal?
        if (the_cat.check_crystal(disabled_crystal))
        {
            return;
        }
        else
        {
            cat_crystal_collisions_disabled = false;
            disabled_crystal = 0;
        } // if ... else
    } // if
    
    for (int i = 0; i < num_crystals; ++i)
    {
        if ( (my_crystals[i].get_row() == row) && (my_crystals[i].is_active()) )
        {
            if ( the_cat.check_crystal(&my_crystals[i]) )
            {
                // make 'ping' sound
                // update player data - i.e. crystals held
                CrystalColor cc = 
                        my_player_data->add_crystal(my_crystals[i].get_color(), 
                                my_play_display);
                redraw_crystals = true;
                my_snd_engine->play_sound(SoundEngine::PING);
                if (cc == C_NONE)
                {
                    my_crystals[i].set_active(false);
                }
                else
                {
                    // Have to stop Albert picking up this crystal 
                    // again straight afterwards...
                    my_crystals[i].set_color(cc);
                    cat_crystal_collisions_disabled = true;
                    disabled_crystal = &my_crystals[i];
                } // if ... else
            } // if
        } // if
    } // for
            
} // FS::SpriteManager::check_cat_crystal_collision()
 
// **************************************************

void FS::SpriteManager::update_logs_and_crystals(const int * current_start_pixels,
                const int * pixels_to_move, const int * old_pixels_to_move)
{
    // update logs
    for (int r = 0; r < NUM_ROWS; ++r)
    {
        for (int c = 0; c < num_logs[r]; ++c)
        {
            my_logs[r][c].update(current_start_pixels[r], 
                    pixels_to_move[r] - old_pixels_to_move[r], is_first_pass,
                    my_play_display);            
        } // for c
    } // for r

    // update crystals if they're active
    for (int i = 0; i < num_crystals; ++i)
    {
        if (my_crystals[i].is_active())
        {
            int r = my_crystals[i].get_row();
            my_crystals[i].update(current_start_pixels[r], 
                    pixels_to_move[r] - old_pixels_to_move[r], is_first_pass,
                    my_play_display);
        } // if
    } // for
    
} // FS::SpriteManager::update_logs_and_crystals()
    
// **************************************************
// **************************************************
// **************************************************
// **************************************************
     
